/**
 * LaBAK Java Cryptography Architecture
 * @author Karol Kubanda
 */
import java.io.*;
import java.security.*;
import java.security.spec.*;
import java.util.*;
import javax.crypto.*;

public class JCALabak {
    public static void main(String[] args) throws IOException, NoSuchAlgorithmException, NoSuchPaddingException, NoSuchProviderException, InvalidAlgorithmParameterException, InvalidKeyException, SignatureException, InvalidKeySpecException {

        //Write available Providers and their Services into file
        FileWriter fw = new FileWriter("ProvidersAndServices.txt");
        Provider[] prvs = Security.getProviders();
        fw.write("Providers and Services:" + "\n");
        for (int i = 0; i < prvs.length; i++) {
            Provider provider = prvs[i];
            fw.write((i + 1) + ". " + provider.getName() + "\n  " + provider.getInfo() + "\n");
            Set<Provider.Service> ser = provider.getServices();
            int counter = 0;
            for (Iterator<Provider.Service> it = ser.iterator(); it.hasNext();) {
                Provider.Service sv = it.next();
                fw.write("     " + (++counter) + ".    " + sv.toString());
            }
        }
        fw.flush(); fw.close();

        //Read file, compute hash, encrypt and write back into file
        MessageDigest sha = MessageDigest.getInstance("SHA-1");
        DigestInputStream din = new DigestInputStream(new FileInputStream("ProvidersAndServices.txt"), sha);

        Cipher tripDES1 = Cipher.getInstance("DESede/CBC/PKCS5Padding");
        SecureRandom random = SecureRandom.getInstance("SHA1PRNG", "SUN");
        KeyGenerator kg = KeyGenerator.getInstance("DESede");
        kg.init(168, random);
        SecretKey sk = kg.generateKey();

        tripDES1.init(Cipher.ENCRYPT_MODE, sk);
        CipherOutputStream cout = new CipherOutputStream(new FileOutputStream("ciphertext.txt"), tripDES1);

        byte[] inBuf = new byte[8192]; int i = 0;
        while ((i = din.read(inBuf)) != -1) {
            cout.write(inBuf, 0, i);
        }
        byte[] md1 = sha.digest();

        //Print tripple des key and iv
        StringBuilder sb = new StringBuilder(); Formatter formatter = new Formatter(sb, Locale.US);
        byte[] key = sk.getEncoded(); byte[] iv = tripDES1.getIV();
        AlgorithmParameters algParams = tripDES1.getParameters();
        for (int s = 0; s < key.length; s++) {
            formatter.format("%1X ", key[s]);
        }
        formatter.format("\n\tIV: ");
        for (int s = 0; s < iv.length; s++) {
            formatter.format("%1X ", iv[s]);
        }
        System.out.println("DESede Key: " + sb);
        cout.flush(); din.close(); cout.close();

        //Read cipher text decrypt, hash and write
        Cipher tripDES2 = Cipher.getInstance("DESede/CBC/PKCS5Padding");
        tripDES2.init(Cipher.DECRYPT_MODE, sk, algParams);
        CipherInputStream cin = new CipherInputStream(new FileInputStream("ciphertext.txt"), tripDES2);

        sha.reset();
        DigestOutputStream dout = new DigestOutputStream(new FileOutputStream("plaintext.txt"), sha);

        i = 0;
        while ((i = cin.read(inBuf)) != -1) {
            dout.write(inBuf, 0, i);
        }
        byte[] md2 = sha.digest();
        dout.flush(); cin.close(); dout.close();

        //print hash
        StringBuilder sb2 = new StringBuilder(); Formatter formatter2 = new Formatter(sb2, Locale.US);
        formatter2.format("Hash: "); boolean equal = true;
        for (int j = 0; j < md2.length; j++) {
            equal &= md1[j]==md2[j];
            formatter2.format("%1X ", md1[j]);
        }
        System.out.println(sb2);
        if(equal) System.out.println("Hashes are equal. "); else System.out.println("Hases are NOT equal!");

        //generate key pair for DSA  1024 and save public and private key in X509 resp, PKCS8 file
        KeyPairGenerator kpg = KeyPairGenerator.getInstance("DSA");
        kpg.initialize(1024);
        KeyPair kp = kpg.genKeyPair();

        KeyFactory kf = KeyFactory.getInstance("DSA");
        X509EncodedKeySpec xeks = kf.getKeySpec(kp.getPublic(), X509EncodedKeySpec.class);
        PKCS8EncodedKeySpec peks = kf.getKeySpec(kp.getPrivate(), PKCS8EncodedKeySpec.class);

        FileOutputStream xout = new FileOutputStream("publickey.der");
        FileOutputStream pout = new FileOutputStream("privatekey.der");
        xout.write(xeks.getEncoded());
        pout.write(peks.getEncoded());
        xout.close(); pout.close();

        //create signature of message.txt file using privatekey file
        FileInputStream fis = new FileInputStream("privatekey.der");
        int read = fis.read(inBuf); fis.close();
        byte[] privKeyBuff = new byte[read];
        System.arraycopy(inBuf, 0, privKeyBuff, 0, read);
        PrivateKey privKey = kf.generatePrivate(new PKCS8EncodedKeySpec(privKeyBuff));

        Signature dsa = Signature.getInstance("SHA1withDSA");
        dsa.initSign(privKey);

        fis = new FileInputStream("message.txt");
        read = fis.read(inBuf); fis.close();

        dsa.update(inBuf, 0, read);
        byte[] sig = dsa.sign();
       
        // verify signature of message txxt using publickey file
        fis = new FileInputStream("publickey.der");
        read = fis.read(inBuf); fis.close();
        byte[] pubKeyBuff = new byte[read];
        System.arraycopy(inBuf, 0, pubKeyBuff, 0, read);
        PublicKey pubKey = kf.generatePublic(new X509EncodedKeySpec(pubKeyBuff));

        dsa.initVerify(pubKey);

        fis = new FileInputStream("message.txt");
        read = fis.read(inBuf); fis.close();
        dsa.update(inBuf, 0, read);
        boolean verifies = dsa.verify(sig);

        System.out.println("Signature correct: " + verifies);
    }
}
